/*
    qZenAstroChoose : Software to easily choose fits pictures in astroPhotography
    Copyright (C) 2017 Olivier Poitevin

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

    Contact : code@zener.fr

*/

#include <math.h>

#include "fitsimage.h"



#include <QFile>
#include <QFileInfo>

FitsImage::FitsImage(QObject *parent) :
    QObject(parent)
{
    // compute color lookup table
    int i;
    for(i=0;i<65536;i++)
    {
        int c=((int)(256*sin((double)i/65536.0*3.1415926/2.0)));
        colCurveLut[i]=c;
        if(c>255)
        {
            colCurveLut[i]=255;
        }
    }
    lastErrMsg=tr("OK");
}

FitsImage::~FitsImage()
{

}

bool FitsImage::loadImage(QString &filename)
{
    Fits_Hdu_st hdu;
    QFile file(filename);

    QFileInfo checkFile(filename);
    if(!checkFile.exists())
    {
        lastErrMsg=tr("File ")+filename+tr(" does not exist");
        return false;
    }

    if(!file.open(QIODevice::ReadOnly))
    {
        lastErrMsg=tr("Cannot open file ")+filename;
        return false;
    }

    QDataStream inStream(&file);
    inStream.readRawData((char*)&hdu,FITSHDUSIZE);

    lastErrMsg=tr("File ")+filename+" : ";
    if(!parseHdu(&hdu))
    {
        file.close();
        return false;
    }

    // Hdu parsed, so
    // load data now
    QByteArray binInput(file.read(width*height*2)); // 2 for 16 bits

    file.close();

    if(binInput.size()!=width*height*2)
    {
        lastErrMsg=tr("File ")+filename+tr("Incorrect size.");
        return false;
    }

    convertCfaToQimg(binInput);

    return true;
}





QImage FitsImage::getImage()
{
    return display;
}

float FitsImage::getExpoTime()
{
    return expoTime;
}

QString FitsImage::getImageDate()
{
    return DateObs;
}

QString FitsImage::getImageTime()
{
    return TimeObs;
}

QString FitsImage::getLastErrMsg()
{
    return lastErrMsg;
}

int FitsImage::getImageWidth()
{
    return width;
}

int FitsImage::getImageHeight()
{
    return height;
}



// parse HDU to get info
bool FitsImage::parseHdu(Fits_Hdu_st *hdu)
{
    int i;

    // foreach line in hdu
    for(i=0;i<FITSHDUSIZE/FITSKEYWORDSIZE;i++)
    {
        char *curHduLine=hdu->keyword[i]; // get current hdu line
        curHduLine[FITSKEYWORDSIZE-1]=0; // null terminated strings !
        QString CurKey=QString::fromUtf8(curHduLine); // get line to Qstring
        QStringList LineSplitted=CurKey.split(" ",QString::SkipEmptyParts); // split by spaces

        // if list has 3 elements => "<KEYWORD> = <value>"
        // if list has 2 elements => "<KEYWORD>= <value>"
        //                        => "<KEYWORD> =  " (empty value)
        // if list has 1 element => "<KEYWORD>= " (empty value)
        // so we have to test 2nd elm first

        switch(LineSplitted.size())
        {
            case 3:
            {
                // remove '='
                LineSplitted.removeAt(1);
                break;
            }
            case 2:
            {
                if(LineSplitted.at(1)=="=")
                {
                    LineSplitted.removeAt(1);
                    LineSplitted.append("");
                }
                else
                {
                    QString Tmp=LineSplitted.at(0);
                    Tmp=Tmp.left(Tmp.length()-1); // cannot be empty
                    LineSplitted.replace(0,Tmp);
                }
                break;
            }
            case 1:
            {
                LineSplitted.append("");
                break;
            }
        }
        QString Tmp=LineSplitted.at(0);
        if(Tmp.length() && Tmp[Tmp.length()-1]=='=')
        {
            Tmp=Tmp.left(Tmp.length()-1);
            LineSplitted.replace(0,Tmp);
        }

        QString KeyName=LineSplitted.at(0);
        QString KeyValue=LineSplitted.at(1);

        if(KeyName==FITKW_END)
        {
            break;
        }

        if(KeyName==FITKW_SIMPLE)
        {
            if(KeyValue!="T")
            {
                lastErrMsg+=tr("Invalid file format (SIMPLE not T).");
                return false;
            }
        }
        else if(KeyName==FITKW_BITPIX)
        {
            if(KeyValue!="16") // only supports 16 bits now
            {
                lastErrMsg+=tr("Invalid file format (supports only 16 bits).");
                return false;
            }
        }
        else if(KeyName==FITKW_NAXISG)
        {
            if(KeyValue!="2") // only supports 2D pictures
            {
                lastErrMsg+=tr("Invalid file format (supports only 2D pictures).");
                return false;
            }
        }
        else if(KeyName==FITKW_NAXIS1)
        {
            bool ok;
            width=KeyValue.toInt(&ok);
            if(!ok)
            {
                lastErrMsg+=tr("Invalid file format (Invalid NAXIS1).");
                return false;
            }
        }
        else if(KeyName==FITKW_NAXIS2)
        {
            bool ok;
            height=KeyValue.toInt(&ok);
            if(!ok)
            {
                lastErrMsg+=tr("Invalid file format (Invalid NAXIS2).");
                return false;
            }
        }
        else if(KeyName==FITKW_MIPSHI)
        {
            bool ok;
            maximumCol=KeyValue.toInt(&ok);
            if(!ok)
            {
                lastErrMsg+=tr("Invalid file format (Invalid MIPSHI).");
                return false;
            }
        }
        else if(KeyName==FITKW_MIPSLO)
        {
            bool ok;
            minimumCol=KeyValue.toInt(&ok);
            if(!ok)
            {
                lastErrMsg+=tr("Invalid file format (Invalid MIPSLO).");
                return false;
            }
        }
        else if(KeyName==FITKW_MIPSX1)
        {
            // must be 0
            continue; // nothing to do now
        }
        else if(KeyName==FITKW_MIPSY1)
        {
            // must be 0
            continue; // nothing to do now
        }
        else if(KeyName==FITKW_MIPSX2)
        {
            // must be 0
            continue; // nothing to do now
        }
        else if(KeyName==FITKW_MIPSY2)
        {
            // must be 0
            continue; // nothing to do now
        }
        else if(KeyName==FITKW_MIPSBIX)
        {
            // must be 1
            continue; // nothing to do now
        }
        else if(KeyName==FITKW_MIPSBIY)
        {
            // must be 1
            continue; // nothing to do now
        }
        else if(KeyName==FITKW_MIPSCCD)
        {
            // must be 0
            continue; // nothing to do now
        }
        else if(KeyName==FITKW_MIPSFL)
        {
            // must be 0
            continue; // nothing to do now
        }
        else if(KeyName==FITKW_MIPSSX)
        {
            // must be same as width
            continue; // nothing to do now
        }
        else if(KeyName==FITKW_MIPSSY)
        {
            // must be same as height
            continue; // nothing to do now
        }
        else if(KeyName==FITKW_MIPSNOR)
        {
            continue; // nothing to do now
        }
        else if(KeyName==FITKW_MIPSCST)
        {
            continue; // nothing to do now
        }
        else if(KeyName==FITKW_DATEOBS)
        {
            DateObs=KeyValue.replace("'","");
        }
        else if(KeyName==FITKW_UTSTART)
        {
            TimeObs=KeyValue.replace("'","");
        }
        else if(KeyName==FITKW_EXPTIME)
        {
            bool ok;
            expoTime=KeyValue.toDouble(&ok);
            if(!ok)
            {
                lastErrMsg+=tr("Invalid file format (Invalid EXPTIME).");
                return false;
            }
        }

    } // end foreach line in hdu

    return true;
}

#define GETPIXVAL(x) ((x>>8)&0xff)|((x<<8)&0xff00)

void FitsImage::convertCfaToQimg(QByteArray &src)
{
    int i;


    int shift=0;
    // find shift order from maximum color
    for(i=0;i<16;i++)
    {
        if(1<<i>maximumCol)
        {
            shift=16-i;
            break;
        }
    }

/*
    CFA will look like this :

        G11 B12 G13 B14

        R21 G22 R23 G24

        G31 B32 G33 B34

        R41 G42 R43 G44


    For each component :

      (-1,-1)  ( 0,-1)  (+1,-1)
      (-1, 0)     C     (+1, 0)
      (-1,+1)  ( 0,+1)  (+1,+1)

This will give :

    RED pix:
    R[(0,0)] G[(0,-1)+(-1,0)+(+1,0)+(0,+1)] B[(-1,-1)+(-1,+1)+(+1,-1)+(+1,+1)]

    GREEN 1 pix (GB line):
    R[(0,-1)+(0,+1)] G[(0,0)] B[(-1,0)+(+1,0)]

    GREEN 2 pix (RG line):
    R[(-1,0)+(+1,0)] G[(0,0)] B[(0,-1)+(0,+1)]

    BLUE pix:
    R[(-1,-1)+(-1,+1)+(+1,-1)+(+1,+1)] G[(0,-1)+(-1,0)+(+1,0)+(0,+1)] B[(0,0)]


    For borders, out-of-bounds pix will be ignored


*/
    display=QImage(QSize(width,height),QImage::Format_RGB32);

    unsigned short *mainDat=(unsigned short*)src.data();


    for(int y=1;y<height-1;y++)
    {
        uint *scanLine=reinterpret_cast<uint *>(display.scanLine(height-y-1));

        unsigned short *dat=&mainDat[y*width+1];
        unsigned short *datUp=&mainDat[(y-1)*width+1];
        unsigned short *datDown=&mainDat[(y+1)*width+1];


        // line
        for(int x=1;x<width-1;x++)
        {
            unsigned int r,g,b;
            unsigned short curPix=*dat;
            unsigned short val;
            unsigned short pup,pdown,pleft,pright;
            unsigned short pul,pur,pdl,pdr;


            pup=GETPIXVAL(*datUp);
            pdown=GETPIXVAL(*datDown);

            pleft=GETPIXVAL(dat[-1]);
            pul=GETPIXVAL(datUp[-1]);
            pdl=GETPIXVAL(datDown[-1]);

            pright=GETPIXVAL(dat[1]);
            pur=GETPIXVAL(datUp[1]);
            pdr=GETPIXVAL(datDown[1]);

            val=GETPIXVAL(curPix);

            dat++;
            datUp++;
            datDown++;

            if(y&0x01)
            {
                if(x&0x01)
                {
                    // GREEN2
                    // R[(-1,0)+(+1,0)] G[(0,0)] B[(0,-1)+(0,+1)]
                    r=(pleft+pright)>>1;
                    g=val;
                    b=(pup+pdown)>>1;
                }
                else
                {
                    // RED
                    // R[(0,0)] G[(0,-1)+(-1,0)+(+1,0)+(0,+1)] B[(-1,-1)+(-1,+1)+(+1,-1)+(+1,+1)]
                    r=val;
                    g=(pup+pleft+pright+pdown)>>2;
                    b=(pul+pur+pdl+pdr)>>2;
                }

            }
            else
            {
                if(x&0x01)
                {
                    // BLUE
                    // R[(-1,-1)+(-1,+1)+(+1,-1)+(+1,+1)] G[(0,-1)+(-1,0)+(+1,0)+(0,+1)] B[(0,0)]
                    r=(pul+pur+pdl+pdr)>>2;
                    g=(pup+pleft+pright+pdown)>>2;
                    b=val;

                }
                else
                {
                    // GREEN1
                    // R[(0,-1)+(0,+1)] G[(0,0)] B[(-1,0)+(+1,0)]
                    r=(pup+pdown)>>1;
                    g=val;
                    b=(pleft+pright)>>1;
                }
            }

            r=colCurveLut[(r<<shift)&0xffff];
            g=colCurveLut[(g<<shift)&0xffff];
            b=colCurveLut[(b<<shift)&0xffff];

            *scanLine++=((r&0xff)<<16)|((g&0xff)<<8)|(b&0xff);

        }
    }


}

