/*
    qZenAstroChoose : Software to easily choose fits pictures in astroPhotography
    Copyright (C) 2017 Olivier Poitevin

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

    Contact : code@zener.fr

*/

#include <QSettings>
#include <QDir>
#include <QShortcut>

#include "mainwindow.h"
#include "ui_mainwindow.h"
#include "dialogpreferences.h"
#include "fitsimage.h"

MainWindow::MainWindow(QWidget *parent) :
    QMainWindow(parent),
    ui(new Ui::MainWindow)
{
    ui->setupUi(this);
    settingsFile="qZenAstroChooseSettings.ini";
    loadSettings();
    curFileIndex=0;
    checkedAsGood=false;
    showStatusMessage(tr("Ready"));

    new QShortcut(QKeySequence(Qt::Key_Right),this,SLOT(on_pushButtonNext_clicked()));
    new QShortcut(QKeySequence(Qt::Key_Left),this,SLOT(on_pushButtonPrevious_clicked()));
    new QShortcut(QKeySequence(Qt::Key_Space),this,SLOT(on_pushButtonKeep_clicked()));
    connect(this,SIGNAL(renderedImage(QImage)),ui->pictureWidget,SLOT(updatePixmap(QImage)));

    on_pushButtonUpdate_clicked();
}

MainWindow::~MainWindow()
{
    saveSettings();
    delete ui;
}

bool MainWindow::loadSettings()
{
    QSettings settings(settingsFile, QSettings::NativeFormat);

    // dirs
    mainDir=settings.value("MainDir","").toString();
    imageDir=settings.value("ImageDir","images").toString();
    imageFilter=settings.value("ImageFilter","lightcfa*.fit").toString();
    // language
    language=settings.value("Language","en").toString();

    return true;
}

bool MainWindow::saveSettings()
{
    QSettings settings(settingsFile, QSettings::NativeFormat);

    // dirs
    settings.setValue("MainDir",mainDir);
    settings.setValue("ImageDir",imageDir);
    settings.setValue("ImageFilter",imageFilter);
    // language
    settings.setValue("Language",language);

    return true;
}

void MainWindow::showStatusMessage(const QString &msg)
{
    ui->statusBar->showMessage(msg);
}

bool MainWindow::getFileList()
{
    QString dirName;
    if(mainDir!="")
    {
        dirName=mainDir+"/"+imageDir;
    }
    else
    {
        dirName=imageDir;
    }
    QDir imageQDir(dirName);
    if(!imageQDir.exists())
    {
        showStatusMessage(tr("Directory ")+dirName+tr(" does not exists"));
        return false;
    }

    QString FileFilterStr;
    if(imageFilter.length())
    {
        FileFilterStr=imageFilter;
    }
    else
    {
        FileFilterStr="*.fit";
    }

    QDir filesQDir(dirName,FileFilterStr);
    QFileInfoList list=filesQDir.entryInfoList();
    if(!list.size())
    {
        showStatusMessage(tr("No fits file found in ")+dirName);
        return false;
    }

    FileList.clear();
    vIsKeep.clear();

    for(int i=0;i<list.size();i++)
    {
        QFileInfo fileInfo=list.at(i);
        if(fileInfo.size())
        {
            FileList<<fileInfo.fileName();
        }
        vIsKeep.append(true);
    }

    if(!FileList.size())
    {
        showStatusMessage(tr("No real fits file found (empty sizes?)"));
        return false;
    }
    curFileIndex=0;
    ui->pushButtonPrevious->setDisabled(true);
    ui->pushButtonNext->setDisabled((FileList.size()>1)?false:true);

    loadKeepFile();

    if(!loadImageFromFits())
    {
        return false;
    }

    return true;
}

bool MainWindow::loadKeepFile()
{
    QString keeFileName;
    if(mainDir!="")
    {
        keeFileName=mainDir+"/"+imageDir+"/"+KEEPFILENAME;
    }
    else
    {
        keeFileName=imageDir+"/"+KEEPFILENAME;
    }
    QSettings keepFile(keeFileName, QSettings::NativeFormat);

    vIsKeep.clear();

    foreach (QString curFile, FileList)
    {
        vIsKeep.append(keepFile.value(curFile,"true").toBool());
    }

    return true;
}

bool MainWindow::saveKeepFile()
{
    QString keeFileName;
    if(mainDir!="")
    {
        keeFileName=mainDir+"/"+imageDir+"/"+KEEPFILENAME;
    }
    else
    {
        keeFileName=imageDir+"/"+KEEPFILENAME;
    }
    QSettings keepFile(keeFileName, QSettings::NativeFormat);

    int i=0;
    foreach (QString curFile, FileList)
    {
        keepFile.setValue(curFile,vIsKeep.at(i++));
    }

    return true;
}

bool MainWindow::loadImageFromFits()
{
    QString fileName;
    if(mainDir!="")
    {
        fileName=mainDir+"/"+imageDir+"/"+FileList.at(curFileIndex);
    }
    else
    {
        fileName=imageDir+"/"+FileList.at(curFileIndex);
    }
    FitsImage fitsImage;

    if(!fitsImage.loadImage(fileName))
    {
        showStatusMessage(fitsImage.getLastErrMsg());
        return false;
    }

    float expoTime=fitsImage.getExpoTime();

    lastImage=fitsImage.getImage();

    QString ExpoTimeStr;
    ExpoTimeStr.sprintf("%g",expoTime);

    showStatusMessage(tr("Image : ")+fileName+" ("+QString::number(curFileIndex+1)+"/"+QString::number(FileList.size())+") Expo : "+ExpoTimeStr+"s"
                      +" when : "+fitsImage.getImageTime()+"-"+fitsImage.getImageDate()
                      +" dim : "+QString::number(fitsImage.getImageWidth())+"x"+QString::number(fitsImage.getImageHeight())
                      );

    checkedAsGood=vIsKeep.at(curFileIndex)?false:true;

    updateKeepButton();
    emit(renderedImage(lastImage));

    return true;
}

void MainWindow::updateKeepButton()
{
    ui->pictureWidget->setDiagPattern(checkedAsGood);
    if(checkedAsGood)
    {
        ui->pushButtonKeep->setText(tr("Keep"));
    }
    else
    {
        ui->pushButtonKeep->setText(tr("Loose"));
    }
}

void MainWindow::on_actionOptions_triggered()
{
    DialogPreferences *dialog=new DialogPreferences(this);

    dialog->setMainDir(mainDir);
    dialog->setImageDir(imageDir);
    dialog->setFilter(imageFilter);
    dialog->setLanguage(language);

    dialog->exec();
    if(dialog->getResult())
    {
        mainDir=dialog->getMainDir();
        imageDir=dialog->getImageDir();
        imageFilter=dialog->getFilter();
        language=dialog->getLanguage();
        on_pushButtonUpdate_clicked();
    }
}

void MainWindow::on_pushButtonUpdate_clicked()
{
    getFileList();
    loadKeepFile();
}

void MainWindow::on_pushButtonPrevious_clicked()
{
    do{
        curFileIndex--;
    }while(curFileIndex>=0 && !loadImageFromFits());

    if(curFileIndex<0)
    {
        curFileIndex=0;
    }
    checkedAsGood=vIsKeep.at(curFileIndex)?false:true;

    ui->pushButtonPrevious->setDisabled(curFileIndex?false:true);
    ui->pushButtonNext->setDisabled((FileList.size()>1)?false:true);

}

void MainWindow::on_pushButtonNext_clicked()
{
    do{
        curFileIndex++;
    }while(curFileIndex<FileList.size() && !loadImageFromFits());
    if(curFileIndex>=FileList.size())
    {
        curFileIndex=FileList.size()-1;
        if(curFileIndex<0)
        {
            curFileIndex=0;
        }
    }
    ui->pushButtonPrevious->setDisabled(curFileIndex?false:true);
    ui->pushButtonNext->setDisabled((curFileIndex<(FileList.size()-1))?false:true);
}

void MainWindow::on_pushButtonKeep_clicked()
{
    checkedAsGood=checkedAsGood?false:true;
    updateKeepButton();

    vIsKeep.replace(curFileIndex,checkedAsGood?false:true);

    saveKeepFile();
}
