/*
    qZenAstroChoose : Software to easily choose fits pictures in astroPhotography
    Copyright (C) 2017 Olivier Poitevin

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

    Contact : code@zener.fr

*/

#include "picturewidget.h"
#include "ui_picturewidget.h"

#include <QtGui>
#include <QtCore>
#include <QScrollArea>
#include <QScrollBar>

PictureWidget::PictureWidget(QWidget *parent) :
    QWidget(parent),
    ui(new Ui::PictureWidget)
{
    pixmap=QPixmap();

    zoomLevel=1.0;
    lastZoomLevel=1.0;
    lastMouseX=0;
    lastMouseY=0;
    zoomMouseX=0;
    zoomMouseX=0;
    justZoomed=false;
    drawDiagPattern=false;
    ui->setupUi(this);
    setSizePolicy(QSizePolicy::Expanding,QSizePolicy::Expanding);

}

PictureWidget::~PictureWidget()
{
    delete ui;
}

void PictureWidget::setDiagPattern(bool val)
{
    drawDiagPattern=val;
    update();
}

void PictureWidget::paintEvent(QPaintEvent *)
{
    QPainter painter(this);

    if (pixmap.isNull())
    {
        painter.fillRect(rect(), Qt::black);
        painter.setPen(Qt::white);
        painter.drawText(rect(), Qt::AlignCenter,tr("NO IMAGE"));
        setBaseSize(0,0);
        setMinimumSize(0,0);
        setSizePolicy(QSizePolicy::Expanding,QSizePolicy::Expanding);
        return;
    }

    setMinimumSize(0,0);
    painter.setRenderHint(QPainter::SmoothPixmapTransform,true);
    painter.drawPixmap(QRect(QPoint(0,0),size()),pixmap,QRect(QPoint(0,0),imageSize));

    // do zoom
    double dispZoomLevel=zoomLevel*zoomLevel;
    QScrollArea *scrollArea=((QScrollArea*)this->parent()->parent());
    setFixedSize(scrollArea->contentsRect().width()*dispZoomLevel,scrollArea->contentsRect().height()*dispZoomLevel);
    setSizePolicy(QSizePolicy::Fixed,QSizePolicy::Fixed);

    // adjust move zoom to follow mouse
    if(justZoomed)
    {
        bool zoomIn=(lastZoomLevel<dispZoomLevel)?true:false;
        QScrollBar *hScroll=scrollArea->horizontalScrollBar();
        QScrollBar *vScroll=scrollArea->verticalScrollBar();

        int hScrVal1=zoomIn?hScroll->value():hScroll->value()*dispZoomLevel/lastZoomLevel;
        int hScrVal2=zoomIn?hScroll->value()*lastZoomLevel/dispZoomLevel:hScroll->value();
        int width=scrollArea->contentsRect().width();
        int leftDisplacement=hScrVal1-hScrVal2;
        int rightDisplacement=hScrVal1-hScrVal2+(zoomIn?width-(int)(width*lastZoomLevel/dispZoomLevel):
                                                        (int)(width*dispZoomLevel/lastZoomLevel)-width);
        int centerVDisplacement=(leftDisplacement+rightDisplacement)/2;
        int mouseXDisplacement=(zoomIn)?zoomMouseX-(int)(zoomMouseX*lastZoomLevel/dispZoomLevel):
                                        (int)(zoomMouseX*dispZoomLevel/lastZoomLevel)-zoomMouseX;
        int finalVDisplacement;
        if(zoomMouseX!=-1)
        {
            finalVDisplacement=((mouseXDisplacement-centerVDisplacement)*2+centerVDisplacement);
        }
        else
        {
            finalVDisplacement=centerVDisplacement;
        }

        int vScrVal1=zoomIn?vScroll->value():vScroll->value()*dispZoomLevel/lastZoomLevel;
        int vScrVal2=zoomIn?vScroll->value()*lastZoomLevel/dispZoomLevel:vScroll->value();
        int height=scrollArea->contentsRect().height();
        int topDisplacement=vScrVal1-vScrVal2;
        int bottomDisplacement=vScrVal1-vScrVal2+(zoomIn?height-(int)(height*lastZoomLevel/dispZoomLevel):
                                                         (int)(height*dispZoomLevel/lastZoomLevel)-height);
        int centerHDisplacement=(topDisplacement+bottomDisplacement)/2;
        int mouseYDisplacement=(zoomIn)?zoomMouseY-(int)(zoomMouseY*lastZoomLevel/dispZoomLevel):
                                        (int)(zoomMouseY*dispZoomLevel/lastZoomLevel)-zoomMouseY;
        int finalHDisplacement;
        if(zoomMouseX!=-1)
        {
            finalHDisplacement=((mouseYDisplacement-centerHDisplacement)*2+centerHDisplacement);
        }
        else
        {
            finalHDisplacement=centerHDisplacement;
        }

        hScroll->setValue(hScroll->value()+finalVDisplacement);
        vScroll->setValue(vScroll->value()+finalHDisplacement);


        justZoomed=false;
        lastZoomLevel=dispZoomLevel;
    }
    if(drawDiagPattern)
    {
        painter.setBrush(QBrush(Qt::red,Qt::FDiagPattern));
        painter.drawRect(0,0,size().width(),size().height());
    }
}

void PictureWidget::updatePixmap(const QImage &image)
{
    if(!image.isNull())
    {
        pixmap = QPixmap::fromImage(image);
        imageSize=image.size();
        if(sizePolicy().horizontalPolicy()!=QSizePolicy::Fixed)
        {
            setBaseSize(imageSize);
        }
    }
    else
    {
        pixmap=QPixmap();
    }
    update();
}

void PictureWidget::wheelEvent(QWheelEvent *event)
{
    if(!pixmap.isNull())
    {
        double newZoomLevel=zoomLevel+(double)event->angleDelta().y()/2000.0;
        if(newZoomLevel>=1.0)
            zoomLevel=newZoomLevel;
        zoomMouseX=event->pos().x();
        zoomMouseY=event->pos().y();

        justZoomed=true;
        update();
    }

    event->accept();
}

void PictureWidget::mousePressEvent(QMouseEvent *event)
{
    if(event->buttons()&Qt::LeftButton)
    {
        // cliking
       lastMouseX=event->pos().x();
       lastMouseY=event->pos().y();
    }
}

void PictureWidget::mouseMoveEvent(QMouseEvent *event)
{
    if(event->buttons()&Qt::LeftButton)
    {
       // moving
       QScrollArea *scrollArea=((QScrollArea*)this->parent()->parent());
       QScrollBar *hScroll=scrollArea->horizontalScrollBar();
       QScrollBar *vScroll=scrollArea->verticalScrollBar();

       hScroll->setValue(hScroll->value()+lastMouseX-event->pos().x());
       vScroll->setValue(vScroll->value()+lastMouseY-event->pos().y());
    }
}

void PictureWidget::doZoom(double inc)
{
    double newZoomLevel=zoomLevel+inc;
    if(newZoomLevel>=1.0)
        zoomLevel=newZoomLevel;
    zoomMouseX=-1;
    zoomMouseY=-1;

    justZoomed=true;
    update();

}

void PictureWidget::zoomIn()
{
    doZoom(0.1);
}

void PictureWidget::zoomOut()
{
    doZoom(-0.1);
}
